/**
 * q-Horn and z-Horn checker
 * Please insert into a.txt your DIMACS CNF problem.
 * It works upto 15 variables.
 * It may have more variables if there are units.
 * If there is more than 15 variables after BCP, 
 * then the checker stops.
 * If you might find an error 
 * please report to the author.
 * There is no guarantee 
 * for the quality of this program!
 * It has been created for research purposes.
 * @author Gabor Kusper, gkusper@aries.ektf.hu
 * @version v1.1. /Eger, Hungary, 02.03.2021/
 * @see http://fmv.ektf.hu/
*/

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

public class qHornzHornChecker {
	public static void main(String[] args) {
		String fileName = "a.txt";
		DIMACSReader reader = new DIMACSReader(fileName);
		ClauseSet cs = new ClauseSet(reader.getN(), reader.getCS());
		System.out.println("The input is:");
		System.out.println(cs);
		ArrayList<Integer> units = reader.getUnits();
		System.out.println("Is there a unit in the input? " );
		System.out.println(units.size() > 0 ? "Yes! We do BCP!" : "No, no unit!" );
		if (units.size() > 0) { 
			ArrayList<Integer> deletedVars = BCP(cs, units); 
			System.out.println("The input after BCP:");
			System.out.println(cs);
			cs.zip(deletedVars);
			System.out.println("The input after ZIP:");
			System.out.println(cs);
		}
		int n = cs.getNumberOfVariables();
		if (reader.getN()>15) {
			System.out.println("This checker is design to check small clause set upto 15 variables!");
			System.out.println("Please, run it on a smaller problem!");
			return;
		}
		double[][] eval = initEval(n);
		/*		
			{
				{0,0,0},{0,0,0.5},{0,0,1},
				{0,0.5,0},{0,0.5,0.5},{0,0.5,1},
				{0,1,0},{0,1,0.5},{0,1,1},
				{0.5,0,0},{0.5,0,0.5},{0.5,0,1},
				{0.5,0.5,0},{0.5,0.5,0.5},{0.5,0.5,1},
				{0.5,1,0},{0.5,1,0.5},{0.5,1,1},
				{1,0,0},{1,0,0.5},{1,0,1},
				{1,0.5,0},{1,0.5,0.5},{1,0.5,1},
				{1,1,0},{1,1,0.5},{1,1,1}
				};
		*/
		QChecker qc = new QChecker();
		double[] qeval = qc.check(cs, eval);
		if (qeval != null) {
			System.out.println("It is a q-Horn clause set!");
			System.out.println("A q-feasible valuation is:");
			for(int i=0; i<qeval.length; i++) {
				System.out.println("beta("+(i+1)+"): " 
									+ qeval[i]);
			}
		}
		else {
			System.out.println("It is NOT a q-Horn clause set!");
		}
		ZChecker zc = new ZChecker();
		double[] zeval = zc.check(cs, eval);
		if (zeval != null) {
			System.out.println("It is a z-Horn clause set!");
			System.out.println("A z-feasible valuation is:");
			for(int i=0; i<zeval.length; i++) {
				System.out.println("gamma("+(i+1)+"): " 
									+ zeval[i]);
			}
		}
		else {
			System.out.println("It is NOT a z-Horn clause set!");
		}
	}
	static ArrayList<Integer> BCP(ClauseSet cs, ArrayList<Integer> units) {
		ArrayList<Integer> deletedVariables = new ArrayList<>();
		while (units.size() > 0) {
			int unit = units.get(0);
			units.remove(0);
			//System.out.println("unit: " + unit);
			deletedVariables.add(Math.abs(unit));
			for(int i=cs.size()-1; i>=0; i--) {
				Clause clause = cs.get(i);
				boolean delete = false;
				int delIndex = -1;
				for(int j=0; j<clause.size(); j++) {
					if (clause.get(j) == unit) { delete = true; break; }
					if (clause.get(j) == -unit) { delIndex = j; break; }
				}
				if (delete) { cs.remove(i); }
				if (delIndex > -1) {
					ArrayList<Integer> literals = new ArrayList<>();
					for(int j=0; j<clause.size(); j++) {
						if (j == delIndex) { continue; }
						literals.add(clause.get(j));
					}
					cs.set(i, new Clause(literals));
					if (literals.size() == 1) {
						int possibleUnit = Math.abs(literals.get(0));
						if (!units.contains(possibleUnit)) { 
							units.add(possibleUnit);
						}
					}
				}
			}
		}
		return deletedVariables;
	}
	
	static double[] valueSet = {0,0.5,1};
	static int[] i;
	static double[][] initEval(int n) {
		//System.out.println("n="+n);
		double[][] eval = new double[(int)Math.pow(3, n)][n];
		if (n==0) return eval;
		i = new int[n];
		i[n-1] = 0;
		initEvalReku(eval, n, n-1);
		i[n-1] = 1;
		initEvalReku(eval, n, n-1);
		i[n-1] = 2;
		initEvalReku(eval, n, n-1);
		return eval;
	}
	static void initEvalReku(
			double[][] eval, int n, int d) {
		if (d==0) {
			createEval(eval, n);
			return;
		}
		//System.out.println("d = " + d);
		i[d-1] = 0;
		initEvalReku(eval, n, d-1);
		i[d-1] = 1;
		initEvalReku(eval, n, d-1);
		i[d-1] = 2;
		initEvalReku(eval, n, d-1);
	}
	static void createEval(double[][] eval, int n) {
		int index = 0;
		for(int j=0; j<n; j++) {
			index *= 3;
			index += i[j];
		}
		//System.out.println("index = " + index);
		for(int j=0; j<n; j++) {
			eval[index][j] = valueSet[i[j]];
		}
	}
}
class QChecker {
	public double[] check(ClauseSet cs, double[][] eval) {
		for(int i=0; i<eval.length; i++) {
			boolean qHorn = true;
			double[] qeval = eval[i];
			for(Clause c : cs.clauses) {
				double sum = 0.0;
				for(int lit : c.literals) {
					int index = lit > 0 ? lit-1 : -lit-1;
					double value = qeval[index];
					sum += lit > 0 ? value : 1-value;
				}
				//System.out.println("sum:" + sum);
				if (sum > 1.0) { qHorn = false; break; }
			}
			if (qHorn) return qeval;
		}
		return null;
	}
}
class ZChecker {
	public double[] check(ClauseSet cs, double[][] eval) {
		for(int i=0; i<eval.length; i++) {
			boolean zHorn = true;
			double[] zeval = eval[i];
			for(Clause c : cs.clauses) {
				//System.out.println("c:" + c);
				double sum = 0.0;
				double half = 0.0;
				for(int lit : c.literals) {
					int lit2 = -lit; 
					int index = lit2 > 0 ? lit2-1 : -lit2-1;
					double value = zeval[index];
					if (value != 0.5) {
						sum += lit2 > 0 ? value : 1-value;
					}
					half += value == 0.5 ? 0.5 : 0;
				}
				//System.out.println("sum:" + sum);
				//System.out.println("half:" + half);
				if (sum < 1.0 && 
						half!=1 &&
						c.literals.size()!=0) { 
					zHorn = false; 
					break; 
				}
			}
			if (zHorn) return zeval;
		}
		return null;
	}
}
class Clause {
	ArrayList<Integer> literals;
	public Clause() { literals = new ArrayList<Integer>(); }
	public Clause(ArrayList<Integer> literals) {
		this.literals = literals;
	}
	public Clause add(int lit) { literals.add(lit); return this; }
	public int size() { return literals.size(); }
	public int get(int i) { return literals.get(i); } 
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for(int i=0; i<literals.size(); i++) {
			sb.append(literals.get(i));
			sb.append(" ");
		}
		sb.append("0");
		return sb.toString();
	}
}
class ClauseSet {
	int numberOfVariables;
	ArrayList<Clause> clauses = new ArrayList<Clause>();
	public int getNumberOfVariables() {
		return numberOfVariables;
	}
	
	public ClauseSet(int n, ArrayList<int[]> cs) {
		numberOfVariables = n;
		for(int[] literals : cs) {
			Clause c = new Clause();
			for(int lit : literals) {
				c.add(lit);
			}
			add(c);
		}
	}
	public void add(Clause clause) { clauses.add(clause); }
	public int size() { return clauses.size(); }
	public Clause get(int i) { return clauses.get(i); }
	public void remove(int i) { clauses.remove(i); }
	public void set(int i, Clause c) { clauses.set(i, c); }
	public void zip(ArrayList<Integer> deletedVars) {
		//System.out.println("deletedVars.size()="+deletedVars.size());
		if (deletedVars.size() == 0) return;
		boolean[] deletedIndeces = 
				new boolean[numberOfVariables+1];
		for(int index : deletedVars) { 
			deletedIndeces[index] = true; 
		}
		int[] decrement = new int[numberOfVariables+1];
		decrement[0] = 0;
		for(int i=1; i<numberOfVariables+1; i++) {
			decrement[i] = decrement[i-1];
			if (deletedIndeces[i]) decrement[i]--;
			//System.out.println("decrement[i]="+decrement[i]);
		}
		numberOfVariables -= deletedVars.size();
		//System.out.println("Let us ZIP!!!");
		for(int i=0; i<clauses.size(); i++) {
			Clause cl = clauses.get(i);
			Clause newCl = new Clause();
			for(int j=0; j<cl.size(); j++) {
				int lit = cl.get(j);
				int par = lit > 0 ? 1 : -1;
				lit = Math.abs(lit);
				lit += decrement[lit];
				lit *= par;
				newCl.add(lit);
			}
			this.set(i, newCl);
		}
	}
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for(int i=0; i<clauses.size(); i++) {
			sb.append(clauses.get(i));
			sb.append("\n");
		}
		return sb.toString();
	}
}
class DIMACSReader {
	private int numberOfVariables;
	private ArrayList<int[]> cs = new ArrayList<int[]>();
	private ArrayList<Integer> units = new ArrayList<Integer>();
	private int[] copyBuffer;
	public ArrayList<int[]> getCS() { return cs; }
	public ArrayList<Integer> getUnits() { return units; }
	public int getN() { return numberOfVariables; }
	
	public DIMACSReader(String fileName) {
		try {
			BufferedReader file = new BufferedReader(new FileReader(fileName));
			String clause = file.readLine();
			while (clause.charAt(0) != 'p') {
				clause = file.readLine();
			}
			readPLine(clause);
			initDataStructure();
			clause = file.readLine();
			while (clause != null) {
				addCNFClause(clause);
				clause = file.readLine();
			}
			file.close();
		} catch (IOException e) { System.out.println(e); System.exit(-1); }
	}
	private void readPLine(String cnfClause) {
		int i1 = 6;
		int i2 = cnfClause.indexOf(" ", i1);
		numberOfVariables = Integer.parseInt(cnfClause.substring(i1, i2));
	}
	private void initDataStructure() {
		copyBuffer = new int[numberOfVariables];
	}
	private void addCNFClause(String cnfClause) {
		if (cnfClause.length() == 0 ||
			cnfClause.charAt(0) == '0' || 
			cnfClause.charAt(0) == 'c' ||
			cnfClause.charAt(0) == '%') { return; }
		int i = 0;
		int i1 = 0;
		cnfClause = cnfClause.replace("\t", " ");
		while(i1 < cnfClause.length())  {
			int i2 = cnfClause.indexOf(" ", i1);
			if (i1 == i2) { i1++; continue; }
			if (i2 == -1) break;
			String lit = cnfClause.substring(i1, i2);
			i1 = i2;
			int literal = Integer.parseInt(lit);
			if (literal == 0) break;
			else copyBuffer[i] = literal;
			i++;
		}
		int[] literals = new int[i];
		for(int index=0; index<i; index++) { 
			literals[index] = copyBuffer[index];
		}
		cs.add(literals);
		if (literals.length == 1 && !units.contains(literals[0])) { 
			units.add(Math.abs(literals[0]));
		}
	}
}